import { get, writable } from 'svelte/store';
function isNullish(value) {
    return value === undefined || value === null;
}
function isEmpty(object) {
    return isNullish(object) || Object.keys(object).length <= 0;
}
const createState = (values, state, validationSchema) => {
    if (validationSchema) {
        state.set(createValidatedState(values, validationSchema.fields));
    }
    else {
        state.set(createRawState(values));
    }
};
const createValidatedState = (data, schema, state = {}) => {
    var _a;
    for (let key in schema) {
        if (schema[key].type === 'array') {
            const values = ((data === null || data === void 0 ? void 0 : data[key]) || []);
            state[key] = values.map((value, index) => { var _a; return createValidatedState(value, schema[key].innerType.fields, Object.assign({}, (_a = state[key]) === null || _a === void 0 ? void 0 : _a[index])); });
            state[key]._errors = [];
        }
        else if (schema[key].type === 'object' && !isEmpty(schema[key].fields)) {
            state[key] = {
                ...createValidatedState(data[key], schema[key].fields, Object.assign({}, state[key])),
                _errors: [],
            };
        }
        else {
            state[key] = {
                _touched: ((_a = state === null || state === void 0 ? void 0 : state[key]) === null || _a === void 0 ? void 0 : _a._touched) || false,
                _errors: [],
            };
        }
    }
    return state;
};
const createRawState = (data, state = {}) => {
    var _a;
    for (let key in data) {
        if (Array.isArray(data[key])) {
            const values = (data[key] || []);
            state[key] = values.map((value, index) => { var _a; return createRawState(value, Object.assign({}, (_a = state[key]) === null || _a === void 0 ? void 0 : _a[index])); });
            state[key]._errors = [];
        }
        else if (typeof data[key] === 'object') {
            state[key] = {
                ...createRawState(data[key], Object.assign({}, state[key])),
                _errors: [],
            };
        }
        else {
            state[key] = {
                _touched: ((_a = state === null || state === void 0 ? void 0 : state[key]) === null || _a === void 0 ? void 0 : _a._touched) || false,
                _errors: [],
            };
        }
    }
    return state;
};
const validate = (form, schema, state, touchValid = false, touchInvalid = false) => {
    try {
        schema.validateSync(get(form), { abortEarly: false });
        clearErrors(state, touchValid);
        return true;
    }
    catch (error) {
        setErrors(state, error, touchValid, touchInvalid);
        return false;
    }
};
const clearErrors = (state, touch = false, inner = false, update = true) => {
    var _a, _b, _c;
    try {
        let data = !inner ? get(state) : state;
        for (let key in data) {
            if (key !== '_touched' && key !== '_errors') {
                if (Array.isArray(data[key])) {
                    let values = (data[key] || []);
                    data[key] = values.map(value => clearErrors(value, touch, true, update));
                    data[key]._touched = typeof ((_a = data[key]) === null || _a === void 0 ? void 0 : _a._touched) === 'boolean' && !touch ? data[key]._touched : touch;
                    data[key]._errors = [];
                }
                else if (typeof data[key] === 'object') {
                    data[key] = {
                        ...clearErrors(data[key], touch, true, update),
                        _touched: typeof ((_b = data[key]) === null || _b === void 0 ? void 0 : _b._touched) === 'boolean' && !touch ? data[key]._touched : touch,
                        _errors: [],
                    };
                }
                else {
                    data[key] = {
                        _touched: typeof ((_c = data[key]) === null || _c === void 0 ? void 0 : _c._touched) === 'boolean' && !touch ? data[key]._touched : touch,
                        _errors: [],
                    };
                }
            }
        }
        if (!inner) {
            update && state.set(data);
            return data;
        }
        else {
            return data;
        }
    }
    catch (error) {
        console.error(error);
        return state;
    }
};
const setErrors = (state, mainError, touchValid = false, touchInvalid = false) => {
    try {
        let _state = get(state);
        _state = clearErrors(_state, touchValid, true, false);
        for (let error of mainError.inner) {
            let segments = error.path.toString().match(/[^.[\]]+/g) || [];
            let obj = _state;
            if (segments.length > 1) {
                const objSegments = segments.slice(0, segments.length - 1);
                let exists = true;
                for (let segment of objSegments) {
                    if (!obj[segment]) {
                        exists = false;
                        break;
                    }
                    obj = obj[segment];
                }
                if (!exists) {
                    break;
                }
            }
            let lastSegment = segments[segments.length - 1];
            if (!obj[lastSegment]) {
                obj[lastSegment] = { touched: false, _errors: [] };
            }
            if (!obj[lastSegment]._errors) {
                obj[lastSegment]._errors = [];
            }
            obj[lastSegment]._errors.push(error.message);
            obj[lastSegment]._touched = typeof obj[lastSegment]._touched === 'boolean' && !touchInvalid ? obj[lastSegment]._touched : touchInvalid;
        }
        state.set(_state);
    }
    catch (error) {
        console.error(error);
    }
};
const updateFieldTouched = (name, state) => {
    let segments = name.toString().match(/[^.[\]]+/g) || [];
    let _state = get(state);
    let obj = _state;
    let exists = true;
    if (segments.length > 1) {
        const objSegments = segments.slice(0, segments.length - 1);
        for (let segment of objSegments) {
            if (!obj[segment]) {
                exists = false;
                break;
            }
            obj = obj[segment];
        }
    }
    if (exists) {
        let lastSegment = segments[segments.length - 1];
        if (obj[lastSegment]) {
            obj[lastSegment]._touched = true;
        }
    }
    state.set(_state);
};
const updateState = (form, validationSchema, state) => {
    let _state = get(state);
    if (validationSchema) {
        _state = createValidatedState(get(form), validationSchema.fields, JSON.parse(JSON.stringify(_state)));
    }
    else {
        _state = createRawState(get(form), JSON.parse(JSON.stringify(_state)));
    }
    state.set(_state);
};
const getFieldState = (name, $state) => {
    let segments = name.toString().match(/[^.[\]]+/g) || [];
    let obj = $state;
    const objSegments = segments.slice(0, segments.length - 1);
    for (let segment of objSegments) {
        if (!obj[segment]) {
            return null;
        }
        obj = obj[segment];
    }
    let lastSegment = segments[segments.length - 1];
    return obj[lastSegment] || null;
};
export function createForm({ values: initialValues, validationSchema, css: cssConfig, validateOnChange, validateOnBlur }) {
    const values = writable(initialValues);
    const state = writable({});
    const isValid = writable(false);
    const isTouched = writable(false);
    const css = {
        ...{
            enabled: true,
            validClass: 'is-valid',
            invalidClass: 'is-invalid',
            useValid: true,
            useInvalid: true,
        }, ...(cssConfig || {})
    };
    validateOnChange = typeof validateOnChange !== 'boolean' ? true : validateOnChange;
    validateOnBlur = typeof validateOnBlur !== 'boolean' ? true : validateOnBlur;
    const cssValidator = writable(0);
    const validationReset = writable(0);
    createState(values, state, validationSchema);
    const updateForm = async () => {
        updateState(values, validationSchema, state);
        validateForm();
    };
    const validateForm = (highlight = 'none') => {
        isValid.set(validate(values, validationSchema, state, highlight === 'all', highlight === 'all' || highlight === 'errors'));
        css.enabled && !validateOnChange && highlight !== 'none' && cssValidator.update(val => val + 1);
    };
    const resetForm = (newValue) => {
        if (newValue) {
            values.set(newValue);
        }
        state.set({});
        isValid.set(false);
        isTouched.set(false);
        createState(get(values), state, validationSchema);
        validationReset.update(val => val + 1);
    };
    values.subscribe(() => {
        validationSchema && validateForm();
    });
    const handleChange = (event) => {
        const target = event === null || event === void 0 ? void 0 : event.target;
        target.name && updateFieldTouched(target.name, state);
        isTouched.set(true);
    };
    const setTouched = (state) => {
        isTouched.set(state);
    };
    const highlight = (node, options = {}) => {
        options = {
            ...{
                enabled: true,
                validClass: 'is-valid',
                invalidClass: 'is-invalid',
                useValid: true,
                useInvalid: true,
            },
            ...options
        };
        console.log(options);
        const changeListener = (event) => {
            if (validateOnChange) {
                handleChange(event);
                if ('enabled' in options ? options.enabled : css.enabled) {
                    checkValidity(get(state));
                }
            }
        };
        const blurListener = (event) => {
            if (validateOnBlur) {
                handleChange(event);
                if ('enabled' in options ? options.enabled : css.enabled) {
                    checkValidity(get(state));
                }
            }
        };
        let unsubscribeState = null;
        let unsubscribeCssValidator = null;
        let unsubscribeValidationReset = null;
        const checkValidity = async ($state) => {
            var _a, _b;
            if (node.name) {
                const fieldState = getFieldState(node.name, $state);
                const invalid = (fieldState === null || fieldState === void 0 ? void 0 : fieldState._touched) && !!((_a = fieldState === null || fieldState === void 0 ? void 0 : fieldState._errors) === null || _a === void 0 ? void 0 : _a.length);
                const valid = (fieldState === null || fieldState === void 0 ? void 0 : fieldState._touched) && !((_b = fieldState === null || fieldState === void 0 ? void 0 : fieldState._errors) === null || _b === void 0 ? void 0 : _b.length);
                if (invalid) {
                    node.classList.remove(options.validClass || css.validClass);
                    if ('useInvalid' in options ? options.useInvalid : css.useInvalid) {
                        node.classList.add(options.invalidClass || css.invalidClass);
                    }
                }
                else if (valid) {
                    node.classList.remove(options.invalidClass || css.invalidClass);
                    if ('useValid' in options ? options.useValid : css.useValid) {
                        node.classList.add(options.validClass || css.validClass);
                    }
                }
            }
        };
        if (['input', 'select', 'textarea'].includes(node.tagName)
            || node.contentEditable) {
            node.addEventListener('change', changeListener);
            node.addEventListener('blur', blurListener);
            if (node.name) {
                unsubscribeState = state.subscribe($state => {
                    ('enabled' in options ? options.enabled : css.enabled) && validateOnChange && checkValidity($state);
                });
                unsubscribeCssValidator = cssValidator.subscribe(() => {
                    checkValidity(get(state));
                });
                unsubscribeValidationReset = validationReset.subscribe(() => {
                    node.classList.remove(options.validClass || css.validClass, options.invalidClass || css.invalidClass);
                });
            }
        }
        return {
            destroy() {
                node.removeEventListener('change', changeListener);
                node.removeEventListener('blur', blurListener);
                unsubscribeState && unsubscribeState();
                unsubscribeCssValidator && unsubscribeCssValidator();
                unsubscribeValidationReset && unsubscribeValidationReset();
            }
        };
    };
    return {
        values,
        state,
        isValid,
        isTouched,
        validateForm,
        handleChange,
        updateForm,
        setTouched,
        highlight,
        resetForm
    };
}
